#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-compat.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <string.h>

/* type for 0-255 */
typedef unsigned char u8;

/* natural type for unsigned integer operations (must be at least 32 bits) */
typedef unsigned long word;

/* cube representation: A cube is represented by the orientation of
 * its 20 movable cubies.  (No whole-cube moves are allowed, so the
 * face center cubies never move.)  The orientation of a cubie is
 * given by the symmetry operator that generates that orientation when
 * applied to the default cubie orientation (the orientation of the
 * cubies in the solved cube).  The cubie positions are as follows:
 * (F = front, B = back, U = up, D = down, L = left, R = right)
 * 0  - FUL corner
 * 1  - FU edge
 * 2  - FUR corner
 * 3  - FL edge
 * 4  - FR edge
 * 5  - FDL corner
 * 6  - FD edge
 * 7  - FDR corner
 * 8  - UL edge
 * 9  - UR edge
 * 10 - DL edge
 * 11 - DR edge
 * 12 - UBL corner
 * 13 - UB edge
 * 14 - UBR corner
 * 15 - BL edge
 * 16 - BR edge
 * 17 - BDL corner
 * 18 - BD edge
 * 19 - BDR corner
 */
typedef struct
{
  word cube[20];                     /* orientation of each cubie */
} cube_type;


/**** table lookup macros ****/

/* A table is an array of bits.  A table is arranged into entries,
 * which are fixed-size chunks of bits.  An entry in a table is
 * indexed by a key.
 * e.g., a four-bit table looks like
 *
 * bits   0001 1011 0011 0101 1111 0000 ...
 * key    0    1    2    3    4    5    ...
 *
 * the functions below extract the machine word holding the given
 * entry, specified by the key, and the offset within the machine word
 * that the given entry resides.
 * 
 * logbits is the logarithm (base 2) of the number of bits in an entry.
 * These functions are designed to inline well when logbits is a constant.
 */
static inline word *get_entry( void *table, word key, word logbits )
{
  return table + ((key >> (3-logbits)) & ~(sizeof(word)-1));
}
static inline word get_offset( word key, word logbits )
{
  return ( key & ((sizeof(word) * 8 - 1) >> logbits) ) << logbits;
}

/* rtab.cilk */
/* table of random numbers used for hashing */
extern unsigned rnd_table[1024];

/* move.cilk */
/*
 *  Moves.
 *  The low order bit encodes clockwise/counterclockwise rotation.
 *  All but the low order two bits specify rotation axis (x,y, or z).
 */
#define      FF  0  /* front forward or clockwise (looking toward center) */
#define      FR  1  /* front reverse or counter (looking toward center)   */
#define      BF  2  /* back  forward */
#define      BR  3  /* back  reverse */
#define      LF  4  /* left  forward */
#define      LR  5  /* left  reverse */
#define      RF  6  /* right forward */
#define      RR  7  /* right reverse */
#define      UF  8  /* up    forward */
#define      UR  9  /* up    reverse */
#define      DF 10  /* down  forward */
#define      DR 11  /* down  reverse */

void solved_cube( cube_type *x );
void scramble_cube( cube_type *x, word n );
void display_cube( cube_type *x );
void move( cube_type *f, cube_type *t, word mv );
void set_cube( cube_type *x, const char *s );
void check_solvability( cube_type *x );
int apply_moves( cube_type *x, int argc, char *argv[] );

/* symm.cilk */
/*
 *  Symmetries
 */
#define ID              0
#define FACE_U_CW       1
#define FACE_U_CCW      2
#define FACE_U_180      3
#define FACE_L_CW       4
#define FACE_L_CCW      5
#define FACE_L_180      6
#define FACE_F_CW       7
#define FACE_F_CCW      8
#define FACE_F_180      9
#define CORNER_0_CW    10
#define CORNER_0_CCW   11
#define CORNER_2_CW    12
#define CORNER_2_CCW   13
#define CORNER_5_CW    14
#define CORNER_5_CCW   15
#define CORNER_7_CW    16
#define CORNER_7_CCW   17
#define EDGE_1         18
#define EDGE_3         19
#define EDGE_4         20
#define EDGE_6         21
#define EDGE_8         22
#define EDGE_9         23

#define INVERT         24
/* form other inverted symmetries using "INVERT +", e.g. "INVERT + EDGE_1" */

/* symm.cilk */
void init_symmetries( void );
void make_symm( cube_type *f, cube_type *t, word symm );
void canonicalize( cube_type *f, cube_type *t, word key );
extern word mult[48][48];
extern word inv[48];
extern word conj[48][48];

/* twobytwo.cilk */
void init_222( void );
void read_222( void );
word hash_222( cube_type *p );

/* canon.cilk */
void init_canon( void );
void read_canon( void );

/* hash.cilk */
/* HASH_TABLE_SIZE (in bytes) must be a power of two.  The larger the
 * table, the faster the code runs.  HASH_TABLE_DEPTH specifies the
 * deepest cubes entered in the table, which should be approximately
 * log_8(HASH_TABLE_SIZE)+2.
 */

/* 1 Gig hash table (maximum size) */
#if 1
#define HASH_TABLE_DEPTH   12
#define HASH_TABLE_SIZE    (1024UL*1024UL*1024UL)
#endif

/* 128 Meg hash table */
#if 0
#define HASH_TABLE_DEPTH   11
#define HASH_TABLE_SIZE    (128UL*1024UL*1024UL)
#endif

#define HASH_TABLE_ENTRIES (HASH_TABLE_SIZE*4)

void init_hash( void );
void read_hash( void );
word hash( cube_type *p );

/******** global tables *********/

/* 4-bit table containing depths of all corner cubes.
 * The key for this table should be generated by hash_222().
 */
extern void *twotable;

/* 8-bit table containing canonicalizing operators (symmetry operators
 * which, when applied to the indexed cube, bring each corner cube
 * into its canonical representation).  This table contains a dummy
 * value if the canonicalizing operator is not unique.
 * The key for this table should be generated by hash_222().
 */

extern void *canontable;

/* 2-bit table containing lower bounds on depths of cubes whose
 * canonical representative hashes to each entry.
 *
 * entry  meaning
 * -----  -------
 *    11  no cubes with depth HASH_TABLE_DEPTH or less
 *    10  no cubes with depth less than HASH_TABLE_DEPTH
 *    01  no cubes with depth less than HASH_TABLE_DEPTH - 1
 *    00  no lower bound known
 *
 * The key for this table should be generated by hash().
 */
extern void *hashtable;

